"""
Since the VAE posteriors are all normal distributions, we can compute an analytical estimate of the evidence.

Analytical Marginal Likelihood
------------------------------

Let:
  - d = number of dimensions
  - θ = (θ₁, θ₂, …, θ_d) is the parameter vector

Priors:
  For each i = 1, …, d:
    p(θᵢ) = 𝒩(0, 1)
          = 1/√(2π) * exp( - θᵢ² / 2 )

Posterior:
  For each i = 1, …, d:
    p(θᵢ | D) = 𝒩(μᵢ, σᵢ²)

Marginal Likelihood (Evidence):
  p(D) = p(D | θ) * p(θ) / p(θ | D)


Using the Laplace approximation, where we assume that the posterior
is approximately Gaussian around its mode, we
evaluate at θ = μ, the posterior mean.


1. Prior at θ = θr:
     p(μ) = ∏[i=1 to d] (1/√(2π) * exp( - θrᵢ² / 2 ))
          = (2π)^(-d/2) * exp( - (1/2) * Σ[i=1 to d] θrᵢ² )

2. Posterior at θ = θr:
     p(μ | D) = ∏[i=1 to d] (1 / (√(2π) * σᵢ))
              = (2π)^(-d/2) * ∏[i=1 to d] (1/σᵢ)

3. Combining to Get p(D):
     p(D) = p(D | θr) * (p(θr) / p(θr | D))
          = p(D | θr) * { (2π)^(-d/2) * exp( - (1/2) * Σ[i=1 to d] θrᵢ² ) }
                      / { (2π)^(-d/2) * ∏[i=1 to d] (1/σᵢ) }
          = p(D | θr) * exp( - (1/2) * Σ[i=1 to d] θrᵢ² ) * ∏[i=1 to d] σᵢ

Log-Domain Expression:
     log p(D) = log p(D | θr)
              - (1/2) * Σ[i=1 to d] θrᵢ²
              + Σ[i=1 to d] log(σᵢ)
"""

import numpy as np


def compute_gaussian_approx_evidence(
    ref_sample: np.ndarray,
    lnl_ref: float,
    posterior_samples: np.ndarray,
    n_bootstraps: int = 10,
    bootstrap_frac: float = 0.8,
):
    """
    Compute the Gaussian Approximation to the log marginal likelihood (ln Z)
    using the MAP estimate.

    Parameters:
    - LnL_map : float
         Log-likelihood evaluated at the MAP estimate.
    - ref_sample : array-like of shape (d,)
         The MAP estimate of the parameters.
    - posterior_samples : np.ndarray of shape (n_samples, d)
         Posterior samples used to compute the standard deviations.

    Returns:
    - lnZ : float
         The estimated log marginal likelihood.

    Formula:
      lnZ = LnL_map - 0.5 * sum(theta_map^2) + sum(log(sigma)),
      where sigma is computed across the posterior samples.
    """

    nsamp, ndim = posterior_samples.shape
    assert ref_sample.shape == (
        ndim,
    ), f"Shape mismatch between ref_sample ({ref_sample.shape}) and posterior_samples dim ({ndim, })"

    args = (ref_sample, lnl_ref, posterior_samples, bootstrap_frac)
    lnz = np.array([_compute_lnz(*args) for _ in range(n_bootstraps)])
    return np.mean(lnz), np.std(lnz)


def _compute_lnz(
    ref_sample: np.ndarray,
    lnl_ref: float,
    posterior_samples: np.ndarray,
    frac: float,
) -> float:
    nsamp, ndim = posterior_samples.shape
    idxs = np.random.choice(
        np.arange(nsamp), size=int(frac * nsamp), replace=False
    )
    samples = posterior_samples[idxs]

    # Compute standard deviation for each parameter (sigma)
    sigma = np.std(samples, axis=0)
    assert (
        sigma.shape == ref_sample.shape
    ), f"Shape mismatch between sigma ({sigma.shape} and ref_sample {ref_sample.shape})"

    # Compute the quadratic penalty term (0.5 * sum(theta_map^2))
    penalty = 0.5 * np.sum(np.square(ref_sample))

    # Compute the "volume" term (sum(log(sigma)))
    sigma_term = np.sum(np.log(sigma))

    # Combine to get the Gaussian approximation to ln Z
    lnZ = lnl_ref - penalty + sigma_term
    return lnZ



import os

import arviz as az
import corner
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np
from starccato_jax import StarccatoVAE
from starccato_jax.plotting import add_quantiles


def sampler_diagnostic_plots(inf_object: az.InferenceData, outdir: str):
    # Plot the trace plot
    az.plot_trace(inf_object, var_names=["z"])
    plt.savefig(os.path.join(outdir, "trace_plot.png"))
    plt.close()

    # # Plot the corner plot
    # corner.corner(inf_object, var_names=["z"])
    # plt.savefig(os.path.join(outdir, "corner_plot.png"))
    # plt.close()


def plot_ci(
    y_obs: jnp.ndarray,
    z_posterior: jnp.ndarray,
    starccato_vae: StarccatoVAE,
    nsamps=100,
    fname=None,
):
    z_samples = z_posterior[
        np.random.choice(z_posterior.shape[0], nsamps, replace=False)
    ]
    y_preds = starccato_vae.generate(z=z_samples)
    y_recon = starccato_vae.reconstruct(y_obs, n_reps=nsamps)

    ypred_qtls = np.quantile(y_preds, [0.025, 0.5, 0.975], axis=0)
    yrecn_qtls = np.quantile(y_recon, [0.025, 0.5, 0.975], axis=0)

    plt.figure(figsize=(5, 3.5))
    plt.plot(y_obs, label="Observed", color="black", lw=2, zorder=-1)
    ax = plt.gca()
    add_quantiles(ax, ypred_qtls, "Posterior", "tab:orange")
    add_quantiles(ax, yrecn_qtls, "Reconstruction", "tab:gray")

    plt.legend(frameon=False)
    if fname is not None:
        plt.savefig(fname)



from typing import Tuple

from jax import numpy as jnp
from jax.random import PRNGKey


def compute_gss_evidence(
    ln_likes: jnp.ndarray, betas: jnp.ndarray, outdir: str, rng: PRNGKey
) -> Tuple[float, float]:
    pass



from .gaussian_approximation_evidence import compute_gaussian_approx_evidence
from .gss_evidence import compute_gss_evidence
from .harmonic_mean_evidence import compute_harmonic_mean_evidence
from .stepping_stone_evidence import compute_stepping_stone_evidence



import harmonic as hm
import jax.numpy as jnp
import matplotlib.pyplot as plt
import numpy as np


def compute_harmonic_mean_evidence(chains: np.ndarray, lnprob: np.ndarray):
    nchains, nsamples, ndim = chains.shape
    assert lnprob.shape == (
        nchains,
        nsamples,
    ), f"lnprob should have shape (nchains={nchains}, nsamples={nsamples}), got {lnprob.shape}"

    # Instantiate harmonic's chains class
    hm_chains = hm.Chains(ndim)
    hm_chains.add_chains_3d(chains, lnprob)

    # Split the chains into the ones which will be used to train the machine learning model and for inference
    chains_train, chains_infer = hm.utils.split_data(
        hm_chains, training_proportion=0.5
    )

    # Select a machine learning model and train it
    model = hm.model.RealNVPModel(ndim, standardize=True, temperature=0.8)
    model.fit(chains_train.samples, verbose=True, epochs=20)

    # samples = chains.reshape((-1, ndim))
    # samp_num = samples.shape[0]
    # flow_samples = model.sample(samp_num)
    # hm.utils.plot_getdist_compare(samples, flow_samples)
    # plt.show()

    # Instantiate harmonic's evidence class
    ev = hm.Evidence(chains_infer.nchains, model)

    # Pass the evidence class the inference chains and compute the evidence!
    ev.add_chains(chains_infer)
    return ev.compute_ln_evidence()



import os
from time import process_time
from typing import Dict

import arviz as az
import jax.numpy as jnp
import jax.random as random
import numpy as np
from numpyro.infer import MCMC, NUTS
from starccato_jax import StarccatoVAE
from tqdm.auto import tqdm

from .core import _run_mcmc
from .evidence import (
    compute_gaussian_approx_evidence,
    compute_stepping_stone_evidence,
)
from .plotting import plot_ci, sampler_diagnostic_plots
from .utils import beta_spaced_samples


def sample(
    data: jnp.ndarray,
    model_path: str = None,
    rng_int: int = 0,
    outdir="out_mcmc",
    num_warmup=500,
    num_samples=1000,
    num_chains=1,
    stepping_stone_lnz: bool = False,
    verbose=True,
    **lnz_kwargs,
) -> MCMC:
    """
    Sample latent variables given the data.

    :param data: Data to condition the latent variables on.
    :param rng: Random number generator.
    :param latent_dim: Dimension of the latent space.
    :return: Sampled latent variables.
    """

    rng = random.PRNGKey(rng_int)
    starccato_vae = StarccatoVAE(model_path)
    mcmc_kwgs = dict(
        y_obs=data,
        starccato_vae=starccato_vae,
        num_samples=num_samples,
        num_warmup=num_warmup,
        rng=rng,
    )

    t0 = process_time()
    mcmc = _run_mcmc(
        **mcmc_kwgs, beta=1.0, num_chains=num_chains, progress_bar=verbose
    )
    inf_object = az.from_numpyro(mcmc)
    t1 = process_time()
    inf_object.sample_stats["sampling_runtime"] = t1 - t0

    _add_gaussian_approx_lnz(inf_object, **lnz_kwargs)

    if stepping_stone_lnz:
        _add_stepping_stone_lnz(mcmc_kwgs, outdir, inf_object, **lnz_kwargs)

    os.makedirs(outdir, exist_ok=True)

    if verbose:
        print(az.summary(inf_object, var_names=["z"]))
        sampler_diagnostic_plots(inf_object, outdir)
    z_samples = inf_object.posterior["z"].values.reshape(
        -1, starccato_vae.latent_dim
    )
    plot_ci(
        data,
        z_samples,
        starccato_vae,
        fname=os.path.join(outdir, "ci_plot.png"),
    )

    inf_object.to_netcdf(os.path.join(outdir, "inference.nc"))

    return mcmc


def _add_gaussian_approx_lnz(inf_object: az.InferenceData, **lnz_kwargs: Dict):
    """Adds Gaussian Approx LnZ to the inference object."""
    zpost = inf_object.posterior["z"].stack(sample=("chain", "draw")).values.T
    lnl = (
        inf_object.log_likelihood.stack(sample=("chain", "draw"))
        .to_array()
        .values.T
    )

    # Shapes --> zpost: (num_samples, latent_dim), lnl: (num_samples,)
    assert zpost.shape[0] == lnl.shape[0]  # same number of samples
    map_idx = np.argmax(lnl)
    t0 = process_time()
    lnz, lnz_uncertainty = compute_gaussian_approx_evidence(
        lnl_ref=lnl[map_idx].ravel(),
        ref_sample=zpost[map_idx].ravel(),
        posterior_samples=zpost,
        n_bootstraps=lnz_kwargs.get("n_bootstraps", 10),
    )
    t1 = process_time()
    runtime = t1 - t0  # in seconds
    print(
        f"GA Log Z: {lnz:.3e} +/- {lnz_uncertainty:.3e} [Runtime: {runtime:.3f} s]"
    )
    # add these to the inference object
    inf_object.sample_stats["gauss_lnz"] = lnz
    inf_object.sample_stats["gauss_lnz_uncertainty"] = lnz_uncertainty
    inf_object.sample_stats["gauss_runtime"] = runtime


def _add_stepping_stone_lnz(
    mcmc_kwgs: Dict,
    outdir: str,
    inf_object: az.InferenceData,
    **lnz_kwargs: Dict,
) -> None:
    """Adds Stepping Stone LnZ to the inference object."""
    num_temps = lnz_kwargs.get("num_temps", 16)
    tempering_betas = beta_spaced_samples(
        num_temps,
        0.3,
        1,
    )
    tempered_lnls = np.zeros((mcmc_kwgs["num_samples"], num_temps))
    t0 = process_time()
    for i in tqdm(range(num_temps), "Tempering"):
        temp_mcmc = _run_mcmc(
            **mcmc_kwgs,
            beta=tempering_betas[i],
            num_chains=1,
            progress_bar=False,
        )
        tempered_lnls[:, i] = temp_mcmc.get_samples()["untempered_loglike"]

    log_z, log_z_uncertainty = compute_stepping_stone_evidence(
        tempered_lnls, tempering_betas, outdir, mcmc_kwgs["rng"]
    )
    t1 = process_time()
    runtime = t1 - t0  # in seconds

    print(
        f"SS Log Z: {log_z:.3e} +/- {log_z_uncertainty:.3e} [Runtime: {runtime:.3f} s]"
    )
    # add these to the inference object
    inf_object.sample_stats["ss_lnz"] = log_z
    inf_object.sample_stats["ss_lnz_uncertainty"] = log_z_uncertainty
    inf_object.sample_stats["ss_runtime"] = runtime
